/*
 * Copyright (C) 2011 Atmark Techno, Inc. All Rights Reserved.
 */

/*
 * The code contained herein is licensed under the GNU General Public
 * License. You may obtain a copy of the GNU General Public License
 * Version 2 or later at the following locations:
 *
 * http://www.opensource.org/licenses/gpl-license.html
 * http://www.gnu.org/copyleft/gpl.html
 */

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/gpio.h>
#include <linux/delay.h>
#include <linux/mx25_fec.h>
#include <mach/common.h>
#include "armadillo400_iomux.h"
#include "board.h"
#include "mx25_pins.h"
#include "devices.h"

static struct pad_desc armadillo400_fec_pads[] = {
	MX25_PAD_FEC_TX_CLK__FEC_TX_CLK(PAD_CTL_HYS | PAD_CTL_PUE | PAD_CTL_PKE | PAD_CTL_PUS_100K_UP),
	MX25_PAD_FEC_RX_DV__FEC_RX_DV(PAD_CTL_PUE),
	MX25_PAD_FEC_RDATA0__FEC_RDATA0(PAD_CTL_PUE),
	MX25_PAD_FEC_TDATA0__FEC_TDATA0(PAD_CTL_PUE | PAD_CTL_PUS_100K_UP | PAD_CTL_SRE_FAST),
	MX25_PAD_FEC_TX_EN__FEC_TX_EN(PAD_CTL_PUE | PAD_CTL_PUS_100K_UP | PAD_CTL_SRE_FAST),
	MX25_PAD_FEC_MDC__FEC_MDC(PAD_CTL_PUE | PAD_CTL_PUS_100K_UP),
	MX25_PAD_FEC_MDIO__FEC_MDIO(PAD_CTL_HYS | PAD_CTL_PUE | PAD_CTL_PUS_22K_UP),
	MX25_PAD_FEC_RDATA1__FEC_RDATA1(PAD_CTL_PUE),
	MX25_PAD_FEC_TDATA1__FEC_TDATA1(PAD_CTL_PUE | PAD_CTL_PUS_100K_UP | PAD_CTL_SRE_FAST),
	MX25_PAD_CTL_GRP_DSE_FEC(PAD_CTL_DSE_LOW),

	/* PHY_RST */
	MX25_PAD_VSTBY_ACK__GPIO_3_18(0),
	/* LINK_LED */
	MX25_PAD_UPLL_BYPCLK__GPIO_3_16(),
};

static void platform_fec_link_led_active(int active)
{
	gpio_direction_output(FEC_LINK_LED_GPIO, !active);
}

static void platform_fec_phy_reset(void)
{
	/* PHY spec says 100us min and Armadillo-400 has low pass filter
	 * for PHY_RST pin. */
	gpio_direction_output(FEC_PHY_RST_GPIO, 0);
	mdelay(1);
	gpio_direction_output(FEC_PHY_RST_GPIO, 1);
}

static void platform_fec_phy_stop(void)
{
	gpio_direction_output(FEC_PHY_RST_GPIO, 0);
}

static void platform_fec_init(struct device *dev)
{
	mxc_iomux_v3a_setup_multiple_pads(armadillo400_fec_pads);

	/*
	 * Set up the GPIO pin.
	 *
	 * PHY_RST: gpio3[18] is ALT 5 mode of pin VSTBY_ACK
	 * LINK_LED: gpio3[16] is ALT 5 mode of pin UPLL_BYPCLK
	 */
	gpio_request(FEC_PHY_RST_GPIO, "PHY_RST");
	gpio_request(FEC_LINK_LED_GPIO, "LINK_LED");

	platform_fec_phy_reset();
	platform_fec_link_led_active(0);
}

static void platform_fec_exit(struct device *dev)
{
	platform_fec_link_led_active(0);
	platform_fec_phy_stop();

	gpio_free(FEC_PHY_RST_GPIO);
	gpio_free(FEC_LINK_LED_GPIO);
}

static int platform_fec_suspend(struct device *dev)
{
	platform_fec_link_led_active(0);

	return 0;
}

static int platform_fec_resume(struct device *dev)
{
	platform_fec_phy_reset();

	return 0;
}

static struct fec_platform_data armadillo400_fec_data = {
	.init = platform_fec_init,
	.exit = platform_fec_exit,
	.suspend = platform_fec_suspend,
	.resume = platform_fec_resume,
	.link_led_active = platform_fec_link_led_active,
};

void __init armadillo400_fec_init(void)
{
	mxc_register_device(&mx25_fec_device, &armadillo400_fec_data);
}
